@@ -5,8 +5,8 @@
 
 #include <errno.h>
 #include <stdlib.h>
-#include <unistd.h>
 #include <math.h>
+#include <string.h>
 
 #include <girara/datastructures.h>
 #include <girara/utils.h>
@@ -20,6 +20,7 @@
 
 #ifdef G_OS_UNIX
 #include <glib-unix.h>
+#include <gio/gunixinputstream.h>
 #endif
 
 #include "bookmarks.h"
@@ -41,25 +42,39 @@
 #include "adjustment.h"
 #include "dbus-interface.h"
 #include "css-definitions.h"
+#include "synctex.h"
+#include "content-type.h"
 
 typedef struct zathura_document_info_s {
   zathura_t* zathura;
-  const char* path;
-  const char* password;
+  char* path;
+  char* password;
   int page_number;
-  const char* mode;
+  char* mode;
+  char* synctex;
 } zathura_document_info_t;
 
 
 static gboolean document_info_open(gpointer data);
-static void zathura_jumplist_reset_current(zathura_t* zathura);
-static void zathura_jumplist_append_jump(zathura_t* zathura);
-static void zathura_jumplist_save(zathura_t* zathura);
 
 #ifdef G_OS_UNIX
 static gboolean zathura_signal_sigterm(gpointer data);
 #endif
 
+static void
+free_document_info(zathura_document_info_t* document_info)
+{
+  if (document_info == NULL) {
+    return;
+  }
+
+  g_free(document_info->path);
+  g_free(document_info->password);
+  g_free(document_info->mode);
+  g_free(document_info->synctex);
+  g_free(document_info);
+}
+
 /* function implementation */
 zathura_t*
 zathura_create(void)
@@ -79,7 +94,8 @@ zathura_create(void)
   }
 
   /* UI */
-  if ((zathura->ui.session = girara_session_create()) == NULL) {
+  zathura->ui.session = girara_session_create();
+  if (zathura->ui.session == NULL) {
     goto error_out;
   }
 
@@ -88,6 +104,9 @@ zathura_create(void)
   zathura->signals.sigterm = g_unix_signal_add(SIGTERM, zathura_signal_sigterm, zathura);
 #endif
 
+  /* MIME type detection */
+  zathura->content_type_context = zathura_content_type_new();
+
   zathura->ui.session->global.data = zathura;
 
   return zathura;
@@ -99,32 +118,27 @@ error_out:
   return NULL;
 }
 
-bool
-zathura_init(zathura_t* zathura)
+static void
+create_directories(zathura_t* zathura)
 {
-  if (zathura == NULL) {
-    return false;
+  static const unsigned int mode = 0711;
+
+  if (g_mkdir_with_parents(zathura->config.config_dir, mode) == -1) {
+    girara_error("Could not create '%s': %s", zathura->config.config_dir,
+                 strerror(errno));
   }
 
-  /* create zathura (config/data) directory */
-  if (g_mkdir_with_parents(zathura->config.config_dir, 0771) == -1) {
-    girara_error("Could not create '%s': %s", zathura->config.config_dir, strerror(errno));
+  if (g_mkdir_with_parents(zathura->config.data_dir, mode) == -1) {
+    girara_error("Could not create '%s': %s", zathura->config.data_dir,
+                 strerror(errno));
   }
+}
 
-  if (g_mkdir_with_parents(zathura->config.data_dir, 0771) == -1) {
-    girara_error("Could not create '%s': %s", zathura->config.data_dir, strerror(errno));
-  }
-
-  /* load plugins */
-  zathura_plugin_manager_load(zathura->plugins.manager);
-
-  /* configuration */
-  config_load_default(zathura);
-  config_load_files(zathura);
-
-  /* UI */
+static bool
+init_ui(zathura_t* zathura)
+{
   if (girara_session_init(zathura->ui.session, "zathura") == false) {
-    goto error_free;
+    return false;
   }
 
   /* girara events */
@@ -132,16 +146,9 @@ zathura_init(zathura_t* zathura)
   zathura->ui.session->events.unknown_command = cb_unknown_command;
 
   /* zathura signals */
-  zathura->signals.refresh_view = g_signal_new("refresh-view",
-                                               GTK_TYPE_WIDGET,
-                                               G_SIGNAL_RUN_LAST,
-                                               0,
-                                               NULL,
-                                               NULL,
-                                               g_cclosure_marshal_generic,
-                                               G_TYPE_NONE,
-                                               1,
-                                               G_TYPE_POINTER);
+  zathura->signals.refresh_view = g_signal_new(
+    "refresh-view", GTK_TYPE_WIDGET, G_SIGNAL_RUN_LAST, 0, NULL, NULL,
+    g_cclosure_marshal_generic, G_TYPE_NONE, 1, G_TYPE_POINTER);
 
   g_signal_connect(G_OBJECT(zathura->ui.session->gtk.view), "refresh-view",
                    G_CALLBACK(cb_refresh_view), zathura);
@@ -151,28 +158,29 @@ zathura_init(zathura_t* zathura)
   gtk_grid_set_row_homogeneous(GTK_GRID(zathura->ui.page_widget), TRUE);
   gtk_grid_set_column_homogeneous(GTK_GRID(zathura->ui.page_widget), TRUE);
   if (zathura->ui.page_widget == NULL) {
-    goto error_free;
+    return false;
   }
 
-  g_signal_connect(G_OBJECT(zathura->ui.session->gtk.window), "size-allocate", G_CALLBACK(cb_view_resized), zathura);
+  g_signal_connect(G_OBJECT(zathura->ui.session->gtk.window), "size-allocate",
+                   G_CALLBACK(cb_view_resized), zathura);
 
   GtkAdjustment* hadjustment = gtk_scrolled_window_get_hadjustment(
                  GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
 
   /* Connect hadjustment signals */
   g_signal_connect(G_OBJECT(hadjustment), "value-changed",
-      G_CALLBACK(cb_view_hadjustment_value_changed), zathura);
+                   G_CALLBACK(cb_view_hadjustment_value_changed), zathura);
   g_signal_connect(G_OBJECT(hadjustment), "changed",
-      G_CALLBACK(cb_view_hadjustment_changed), zathura);
+                   G_CALLBACK(cb_view_hadjustment_changed), zathura);
 
   GtkAdjustment* vadjustment = gtk_scrolled_window_get_vadjustment(
                  GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
 
   /* Connect vadjustment signals */
   g_signal_connect(G_OBJECT(vadjustment), "value-changed",
-      G_CALLBACK(cb_view_vadjustment_value_changed), zathura);
+                   G_CALLBACK(cb_view_vadjustment_value_changed), zathura);
   g_signal_connect(G_OBJECT(vadjustment), "changed",
-      G_CALLBACK(cb_view_vadjustment_changed), zathura);
+                   G_CALLBACK(cb_view_vadjustment_changed), zathura);
 
   /* page view alignment */
   gtk_widget_set_halign(zathura->ui.page_widget, GTK_ALIGN_CENTER);
@@ -186,66 +194,39 @@ zathura_init(zathura_t* zathura)
   gtk_widget_show(zathura->ui.page_widget);
 
   /* statusbar */
-  zathura->ui.statusbar.file = girara_statusbar_item_add(zathura->ui.session, TRUE, TRUE, TRUE, NULL);
+  zathura->ui.statusbar.file =
+    girara_statusbar_item_add(zathura->ui.session, TRUE, TRUE, TRUE, NULL);
   if (zathura->ui.statusbar.file == NULL) {
-    goto error_free;
+    return false;
   }
 
-  zathura->ui.statusbar.buffer = girara_statusbar_item_add(zathura->ui.session, FALSE, FALSE, FALSE, NULL);
+  zathura->ui.statusbar.buffer =
+    girara_statusbar_item_add(zathura->ui.session, FALSE, FALSE, FALSE, NULL);
   if (zathura->ui.statusbar.buffer == NULL) {
-    goto error_free;
+    return false;
   }
 
-  zathura->ui.statusbar.page_number = girara_statusbar_item_add(zathura->ui.session, FALSE, FALSE, FALSE, NULL);
+  zathura->ui.statusbar.page_number =
+    girara_statusbar_item_add(zathura->ui.session, FALSE, FALSE, FALSE, NULL);
   if (zathura->ui.statusbar.page_number == NULL) {
-    goto error_free;
+    return false;
   }
 
-  girara_statusbar_item_set_text(zathura->ui.session, zathura->ui.statusbar.file, _("[No name]"));
+  girara_statusbar_item_set_text(zathura->ui.session,
+                                 zathura->ui.statusbar.file, _("[No name]"));
 
   /* signals */
-  g_signal_connect(G_OBJECT(zathura->ui.session->gtk.window), "destroy", G_CALLBACK(cb_destroy), zathura);
+  g_signal_connect(G_OBJECT(zathura->ui.session->gtk.window), "destroy",
+                   G_CALLBACK(cb_destroy), zathura);
 
-  /* database */
-  char* database = NULL;
-  girara_setting_get(zathura->ui.session, "database", &database);
+  return true;
+}
 
-  if (g_strcmp0(database, "plain") == 0) {
-    girara_debug("Using plain database backend.");
-    zathura->database = zathura_plaindatabase_new(zathura->config.data_dir);
-#ifdef WITH_SQLITE
-  } else if (g_strcmp0(database, "sqlite") == 0) {
-    girara_debug("Using sqlite database backend.");
-    char* tmp = g_build_filename(zathura->config.data_dir, "bookmarks.sqlite", NULL);
-    zathura->database = zathura_sqldatabase_new(tmp);
-    g_free(tmp);
-#endif
-  } else if (g_strcmp0(database, "null") != 0) {
-    girara_error("Database backend '%s' is not supported.", database);
-  }
-
-  if (zathura->database == NULL && g_strcmp0(database, "null") != 0) {
-    girara_error("Unable to initialize database. Bookmarks won't be available.");
-  } else {
-    g_object_set(G_OBJECT(zathura->ui.session->command_history), "io", zathura->database, NULL);
-  }
-  g_free(database);
-
-  /* bookmarks */
-  zathura->bookmarks.bookmarks = girara_sorted_list_new2((girara_compare_function_t) zathura_bookmarks_compare,
-                                 (girara_free_function_t) zathura_bookmark_free);
-
-  /* jumplist */
-  int jumplist_size = 20;
-  girara_setting_get(zathura->ui.session, "jumplist-size", &jumplist_size);
-
-  zathura->jumplist.max_size = jumplist_size < 0 ? 0 : jumplist_size;
-  zathura->jumplist.list = NULL;
-  zathura->jumplist.size = 0;
-  zathura->jumplist.cur = NULL;
-
-  /* CSS for index mode */
-  GiraraTemplate* csstemplate = girara_session_get_template(zathura->ui.session);
+static void
+init_css(zathura_t* zathura)
+{
+  GiraraTemplate* csstemplate =
+    girara_session_get_template(zathura->ui.session);
 
   static const char* index_settings[] = {
     "index-fg",
@@ -257,8 +238,8 @@ zathura_init(zathura_t* zathura)
   for (size_t s = 0; s < LENGTH(index_settings); ++s) {
     girara_template_add_variable(csstemplate, index_settings[s]);
 
-    char* tmp_value = NULL;
-    GdkRGBA rgba = { 0, 0, 0, 0 };
+    char*   tmp_value = NULL;
+    GdkRGBA rgba      = {0, 0, 0, 0};
     girara_setting_get(zathura->ui.session, index_settings[s], &tmp_value);
     if (tmp_value != NULL) {
       gdk_rgba_parse(&rgba, tmp_value);
@@ -266,14 +247,111 @@ zathura_init(zathura_t* zathura)
     }
 
     char* color = gdk_rgba_to_string(&rgba);
-    girara_template_set_variable_value(csstemplate,
-        index_settings[s], color);
+    girara_template_set_variable_value(csstemplate, index_settings[s], color);
     g_free(color);
   }
 
-  char* css = g_strdup_printf("%s\n%s", girara_template_get_base(csstemplate), CSS_TEMPLATE_INDEX);
+  char* css = g_strdup_printf("%s\n%s", girara_template_get_base(csstemplate),
+                              CSS_TEMPLATE_INDEX);
   girara_template_set_base(csstemplate, css);
   g_free(css);
+}
+
+static void
+init_database(zathura_t* zathura)
+{
+  char* database = NULL;
+  girara_setting_get(zathura->ui.session, "database", &database);
+
+  if (g_strcmp0(database, "plain") == 0) {
+    girara_debug("Using plain database backend.");
+    zathura->database = zathura_plaindatabase_new(zathura->config.data_dir);
+#ifdef WITH_SQLITE
+  } else if (g_strcmp0(database, "sqlite") == 0) {
+    girara_debug("Using sqlite database backend.");
+    char* tmp =
+      g_build_filename(zathura->config.data_dir, "bookmarks.sqlite", NULL);
+    zathura->database = zathura_sqldatabase_new(tmp);
+    g_free(tmp);
+#endif
+  } else if (g_strcmp0(database, "null") != 0) {
+    girara_error("Database backend '%s' is not supported.", database);
+  }
+
+  if (zathura->database == NULL && g_strcmp0(database, "null") != 0) {
+    girara_error(
+      "Unable to initialize database. Bookmarks won't be available.");
+  }
+  else {
+    g_object_set(G_OBJECT(zathura->ui.session->command_history), "io",
+                 zathura->database, NULL);
+  }
+  g_free(database);
+}
+
+static void
+init_jumplist(zathura_t* zathura)
+{
+  int jumplist_size = 20;
+  girara_setting_get(zathura->ui.session, "jumplist-size", &jumplist_size);
+
+  zathura->jumplist.max_size = jumplist_size < 0 ? 0 : jumplist_size;
+  zathura->jumplist.list     = NULL;
+  zathura->jumplist.size     = 0;
+  zathura->jumplist.cur      = NULL;
+}
+
+static void
+init_shortcut_helpers(zathura_t* zathura)
+{
+  zathura->shortcut.mouse.x = 0;
+  zathura->shortcut.mouse.y = 0;
+
+  zathura->shortcut.toggle_page_mode.pages = 2;
+
+  zathura->shortcut.toggle_presentation_mode.pages                  = 1;
+  zathura->shortcut.toggle_presentation_mode.first_page_column_list = NULL;
+  zathura->shortcut.toggle_presentation_mode.zoom                   = 1.0;
+}
+
+bool
+zathura_init(zathura_t* zathura)
+{
+  if (zathura == NULL) {
+    return false;
+  }
+
+  /* create zathura (config/data) directory */
+  create_directories(zathura);
+
+  /* load plugins */
+  zathura_plugin_manager_load(zathura->plugins.manager);
+
+  /* configuration */
+  config_load_default(zathura);
+  config_load_files(zathura);
+
+  /* UI */
+  if (!init_ui(zathura)) {
+    goto error_free;
+  }
+
+  /* database */
+  init_database(zathura);
+
+  /* bookmarks */
+  zathura->bookmarks.bookmarks = girara_sorted_list_new2(
+    (girara_compare_function_t)zathura_bookmarks_compare,
+    (girara_free_function_t)zathura_bookmark_free);
+
+  /* jumplist */
+  init_jumplist(zathura);
+
+  /* CSS for index mode */
+  init_css(zathura);
+
+  /* Shortcut helpers */
+  init_shortcut_helpers(zathura);
 
   /* Start D-Bus service */
   bool dbus = true;
@@ -302,6 +380,16 @@ zathura_free(zathura_t* zathura)
 
   document_close(zathura, false);
 
+  /* MIME type detection */
+  zathura_content_type_free(zathura->content_type_context);
+
+#ifdef G_OS_UNIX
+  if (zathura->signals.sigterm > 0) {
+    g_source_remove(zathura->signals.sigterm);
+    zathura->signals.sigterm = 0;
+  }
+#endif
+
   /* stop D-Bus */
   if (zathura->dbus != NULL) {
     g_object_unref(zathura->dbus);
@@ -312,6 +400,11 @@ zathura_free(zathura_t* zathura)
     girara_session_destroy(zathura->ui.session);
   }
 
+  /* shortcut */
+  if (zathura->shortcut.toggle_presentation_mode.first_page_column_list != NULL) {
+    g_free(zathura->shortcut.toggle_presentation_mode.first_page_column_list);
+  }
+
   /* stdin support */
   if (zathura->stdin_support.file != NULL) {
     g_unlink(zathura->stdin_support.file);
@@ -439,11 +532,10 @@ zathura_set_argv(zathura_t* zathura, char** argv)
   zathura->global.arguments = argv;
 }
 
+#ifdef G_OS_UNIX
 static gchar*
-prepare_document_open_from_stdin(zathura_t* zathura, const char* path)
+prepare_document_open_from_stdin(const char* path)
 {
-  g_return_val_if_fail(zathura, NULL);
-
   int infileno = -1;
   if (g_strcmp0(path, "-") == 0) {
     infileno = fileno(stdin);
@@ -458,10 +550,61 @@ prepare_document_open_from_stdin(zathura_t* zathura, const char* path)
     return NULL;
   }
 
-  GError* error = NULL;
-  gchar* file = NULL;
-  gint handle = g_file_open_tmp("zathura.stdin.XXXXXX", &file, &error);
-  if (handle == -1) {
+  if (infileno == -1) {
+    girara_error("Can not read from file descriptor.");
+    return NULL;
+  }
+
+  GInputStream* input_stream = g_unix_input_stream_new(infileno, false);
+  if (input_stream == NULL) {
+    girara_error("Can not read from file descriptor.");
+    return NULL;
+
+  }
+
+  GFileIOStream* iostream = NULL;
+  GError*        error    = NULL;
+  GFile* tmpfile = g_file_new_tmp("zathura.stdin.XXXXXX", &iostream, &error);
+  if (tmpfile == NULL) {
+    if (error != NULL) {
+      girara_error("Can not create temporary file: %s", error->message);
+      g_error_free(error);
+    }
+    g_object_unref(input_stream);
+    return NULL;
+  }
+
+  const ssize_t count = g_output_stream_splice(
+    g_io_stream_get_output_stream(G_IO_STREAM(iostream)), input_stream,
+    G_OUTPUT_STREAM_SPLICE_NONE, NULL, &error);
+  g_object_unref(input_stream);
+  g_object_unref(iostream);
+  if (count == -1) {
+    if (error != NULL) {
+      girara_error("Can not write to temporary file: %s", error->message);
+      g_error_free(error);
+    }
+    g_file_delete(tmpfile, NULL, NULL);
+    g_object_unref(tmpfile);
+    return NULL;
+  }
+
+  char* file = g_file_get_path(tmpfile);
+  g_object_unref(tmpfile);
+
+  return file;
+}
+#endif
+
+static gchar*
+prepare_document_open_from_gfile(GFile* source)
+{
+  gchar*         file     = NULL;
+  GFileIOStream* iostream = NULL;
+  GError*        error    = NULL;
+
+  GFile* tmpfile = g_file_new_tmp("zathura.gio.XXXXXX", &iostream, &error);
+  if (tmpfile == NULL) {
     if (error != NULL) {
       girara_error("Can not create temporary file: %s", error->message);
       g_error_free(error);
@@ -469,36 +612,22 @@ prepare_document_open_from_stdin(zathura_t* zathura, const char* path)
     return NULL;
   }
 
-  // read and dump to temporary file
-  if (infileno == -1) {
-    girara_error("Can not read from file descriptor.");
-    close(handle);
-    g_unlink(file);
-    g_free(file);
-    return NULL;
-  }
-
-  char buffer[BUFSIZ];
-  ssize_t count = 0;
-  while ((count = read(infileno, buffer, BUFSIZ)) > 0) {
-    if (write(handle, buffer, count) != count) {
-      girara_error("Can not write to temporary file: %s", file);
-      close(handle);
-      g_unlink(file);
-      g_free(file);
-      return NULL;
+  gboolean rc = g_file_copy(source, tmpfile, G_FILE_COPY_OVERWRITE, NULL, NULL,
+                            NULL, &error);
+  if (rc == FALSE) {
+    if (error != NULL) {
+      girara_error("Can not copy to temporary file: %s", error->message);
+      g_error_free(error);
     }
-  }
-
-  close(handle);
-
-  if (count != 0) {
-    girara_error("Can not read from file descriptor.");
-    g_unlink(file);
-    g_free(file);
+    g_object_unref(iostream);
+    g_object_unref(tmpfile);
     return NULL;
   }
 
+  file = g_file_get_path(tmpfile);
+  g_object_unref(iostream);
+  g_object_unref(tmpfile);
+
   return file;
 }
 
@@ -507,12 +636,15 @@ document_info_open(gpointer data)
 {
   zathura_document_info_t* document_info = data;
   g_return_val_if_fail(document_info != NULL, FALSE);
+  char* uri = NULL;
 
   if (document_info->zathura != NULL && document_info->path != NULL) {
     char* file = NULL;
     if (g_strcmp0(document_info->path, "-") == 0 ||
         g_str_has_prefix(document_info->path, "/proc/self/fd/") == true) {
-      file = prepare_document_open_from_stdin(document_info->zathura, document_info->path);
+#ifdef G_OS_UNIX
+      file = prepare_document_open_from_stdin(document_info->path);
+#endif
       if (file == NULL) {
         girara_notify(document_info->zathura->ui.session, GIRARA_ERROR,
                       _("Could not read file from stdin and write it to a temporary file."));
@@ -520,13 +652,35 @@ document_info_open(gpointer data)
         document_info->zathura->stdin_support.file = g_strdup(file);
       }
     } else {
-      file = g_strdup(document_info->path);
+      GFile* gf = g_file_new_for_commandline_arg(document_info->path);
+      if (g_file_is_native(gf) == TRUE) {
+        /* file was given as a native path */
+        file = g_file_get_path(gf);
+      }
+      else {
+        /* copy file with GIO */
+        uri = g_file_get_uri(gf);
+        file = prepare_document_open_from_gfile(gf);
+        if (file == NULL) {
+          girara_notify(document_info->zathura->ui.session, GIRARA_ERROR,
+                        _("Could not read file from GIO and copy it to a temporary file."));
+        } else {
+          document_info->zathura->stdin_support.file = g_strdup(file);
+        }
+      }
+      g_object_unref(gf);
     }
 
     if (file != NULL) {
-      document_open(document_info->zathura, file, document_info->password,
-                    document_info->page_number);
+      if (document_info->synctex != NULL) {
+        document_open_synctex(document_info->zathura, file, uri, 
+                              document_info->password, document_info->synctex);
+      } else {
+        document_open(document_info->zathura, file, uri, document_info->password,
+                      document_info->page_number);
+      }
       g_free(file);
+      g_free(uri);
 
       if (document_info->mode != NULL) {
         if (g_strcmp0(document_info->mode, "presentation") == 0) {
@@ -542,12 +696,69 @@ document_info_open(gpointer data)
     }
   }
 
-  g_free(document_info);
+  free_document_info(document_info);
+  return FALSE;
+}
+
+char*
+get_formatted_filename(zathura_t* zathura, bool statusbar)
+{
+  bool basename_only = false;
+  const char* file_path = zathura_document_get_uri(zathura->document);
+  if (file_path == NULL) {
+    file_path = zathura_document_get_path(zathura->document);
+  }
+  if (statusbar == true) {
+    girara_setting_get(zathura->ui.session, "statusbar-basename", &basename_only);
+  } else {
+    girara_setting_get(zathura->ui.session, "window-title-basename", &basename_only);
+  }
+
+  if (basename_only == false) {
+    bool home_tilde = false;
+    if (statusbar) {
+      girara_setting_get(zathura->ui.session, "statusbar-home-tilde", &home_tilde);
+    } else {
+      girara_setting_get(zathura->ui.session, "window-title-home-tilde", &home_tilde);
+    }
+
+    const size_t file_path_len = file_path ? strlen(file_path) : 0;
+
+    if (home_tilde == true) {
+      char* home = girara_get_home_directory(NULL);
+      const size_t home_len = home ? strlen(home) : 0;
+
+      if (home_len > 1
+          && file_path_len >= home_len
+          && g_str_has_prefix(file_path, home)
+          && (!file_path[home_len] || file_path[home_len] == '/')) {
+        g_free(home);
+        return g_strdup_printf("~%s", &file_path[home_len]);
+      } else {
+        g_free(home);
+        return g_strdup(file_path);
+      }
+    } else {
+      return g_strdup(file_path);
+    }
+  } else {
+    const char* basename = zathura_document_get_basename(zathura->document);
+    return g_strdup(basename);
+  }
+}
+
+static gboolean
+document_open_password_dialog(gpointer data)
+{
+  zathura_password_dialog_info_t* password_dialog_info = data;
+
+  girara_dialog(password_dialog_info->zathura->ui.session, _("Enter password:"), true, NULL,
+                (girara_callback_inputbar_activate_t) cb_password_dialog, password_dialog_info);
   return FALSE;
 }
 
 bool
-document_open(zathura_t* zathura, const char* path, const char* password,
+document_open(zathura_t* zathura, const char* path, const char* uri, const char* password,
               int page_number)
 {
   if (zathura == NULL || zathura->plugins.manager == NULL || path == NULL) {
@@ -556,18 +767,19 @@ document_open(zathura_t* zathura, const char* path, const char* password,
 
   gchar* file_uri = NULL;
   zathura_error_t error = ZATHURA_ERROR_OK;
-  zathura_document_t* document = zathura_document_open(zathura->plugins.manager, path, password, &error);
+  zathura_document_t* document = zathura_document_open(zathura, path, uri, password, &error);
 
   if (document == NULL) {
     if (error == ZATHURA_ERROR_INVALID_PASSWORD) {
+      girara_debug("Invalid or no password.");
       zathura_password_dialog_info_t* password_dialog_info = malloc(sizeof(zathura_password_dialog_info_t));
       if (password_dialog_info != NULL) {
         password_dialog_info->zathura = zathura;
+        password_dialog_info->path = g_strdup(path);
+        password_dialog_info->uri = g_strdup(uri);
 
-        if (path != NULL) {
-          password_dialog_info->path = g_strdup(path);
-          girara_dialog(zathura->ui.session, "Enter password:", true, NULL,
-                        (girara_callback_inputbar_activate_t) cb_password_dialog, password_dialog_info);
+        if (password_dialog_info->path != NULL) {
+          gdk_threads_add_idle(document_open_password_dialog, password_dialog_info);
           goto error_out;
         } else {
           free(password_dialog_info);
@@ -580,6 +792,8 @@ document_open(zathura_t* zathura, const char* path, const char* password,
     }
     goto error_out;
   }
+  
+  zathura->document = document;
 
   const char* file_path        = zathura_document_get_path(document);
   unsigned int number_of_pages = zathura_document_get_number_of_pages(document);
@@ -597,7 +811,7 @@ document_open(zathura_t* zathura, const char* path, const char* password,
     .scale = 1,
     .rotation = 0,
     .pages_per_row = 0,
-    .first_page_column = 0,
+    .first_page_column_list = NULL,
     .position_x = 0,
     .position_y = 0
   };
@@ -665,13 +879,9 @@ document_open(zathura_t* zathura, const char* path, const char* password,
   zathura->bisect.end = number_of_pages - 1;
 
   /* update statusbar */
-  bool basename_only = false;
-  girara_setting_get(zathura->ui.session, "statusbar-basename", &basename_only);
-  if (basename_only == false) {
-    girara_statusbar_item_set_text(zathura->ui.session, zathura->ui.statusbar.file, file_path);
-  } else {
-    girara_statusbar_item_set_text(zathura->ui.session, zathura->ui.statusbar.file, zathura_document_get_basename(document));
-  }
+  char* filename = get_formatted_filename(zathura, true);
+  girara_statusbar_item_set_text(zathura->ui.session, zathura->ui.statusbar.file, filename);
+  g_free(filename);
 
   /* install file monitor */
   file_uri = g_filename_to_uri(file_path, NULL, NULL);
@@ -715,8 +925,6 @@ document_open(zathura_t* zathura, const char* path, const char* password,
     goto error_free;
   }
 
-  zathura->document = document;
-
   /* page cache size */
   int cache_size = 0;
   girara_setting_get(zathura->ui.session, "page-cache-size", &cache_size);
@@ -734,7 +942,7 @@ document_open(zathura_t* zathura, const char* path, const char* password,
   }
 
   /* set up recolor info in ZathuraRenderer */
-  char* recolor_dark = NULL;
+  char* recolor_dark  = NULL;
   char* recolor_light = NULL;
   girara_setting_get(zathura->ui.session, "recolor-darkcolor", &recolor_dark);
   girara_setting_get(zathura->ui.session, "recolor-lightcolor", &recolor_light);
@@ -799,7 +1007,7 @@ document_open(zathura_t* zathura, const char* path, const char* password,
 
   /* view mode */
   unsigned int pages_per_row = 1;
-  unsigned int first_page_column = 1;
+  char* first_page_column_list = NULL;
   unsigned int page_padding = 1;
 
   girara_setting_get(zathura->ui.session, "page-padding", &page_padding);
@@ -810,14 +1018,21 @@ document_open(zathura_t* zathura, const char* path, const char* password,
     girara_setting_get(zathura->ui.session, "pages-per-row", &pages_per_row);
   }
 
-  if (file_info.first_page_column > 0) {
-    first_page_column = file_info.first_page_column;
+  /* read first_page_column list */
+  if (file_info.first_page_column_list != NULL && strcmp(file_info.first_page_column_list, "")) {
+    first_page_column_list = file_info.first_page_column_list;
+    file_info.first_page_column_list = NULL;
   } else {
-    girara_setting_get(zathura->ui.session, "first-page-column", &first_page_column);
+    girara_setting_get(zathura->ui.session, "first-page-column", &first_page_column_list);
   }
 
+  /* find value for first_page_column */
+  unsigned int first_page_column = find_first_page_column(first_page_column_list, pages_per_row);
+
   girara_setting_set(zathura->ui.session, "pages-per-row", &pages_per_row);
-  girara_setting_set(zathura->ui.session, "first-page-column", &first_page_column);
+  girara_setting_set(zathura->ui.session, "first-page-column", first_page_column_list);
+  g_free(file_info.first_page_column_list);
+  g_free(first_page_column_list);
 
   page_widget_set_mode(zathura, page_padding, pages_per_row, first_page_column);
   zathura_document_set_page_layout(zathura->document, page_padding, pages_per_row, first_page_column);
@@ -838,13 +1053,9 @@ document_open(zathura_t* zathura, const char* path, const char* password,
   }
 
   /* update title */
-  basename_only = false;
-  girara_setting_get(zathura->ui.session, "window-title-basename", &basename_only);
-  if (basename_only == false) {
-    girara_set_window_title(zathura->ui.session, file_path);
-  } else {
-    girara_set_window_title(zathura->ui.session, zathura_document_get_basename(document));
-  }
+  char* formatted_filename = get_formatted_filename(zathura, false);
+  girara_set_window_title(zathura->ui.session, formatted_filename);
+  g_free(formatted_filename);
 
   g_free(file_uri);
 
@@ -894,13 +1105,37 @@ error_out:
   return false;
 }
 
+bool
+document_open_synctex(zathura_t* zathura, const char* path, const char* uri,
+                      const char* password, const char* synctex)
+{
+  bool ret = document_open(zathura, path, password, uri,
+                           ZATHURA_PAGE_NUMBER_UNSPECIFIED);
+  if (ret == false) {
+    return false;
+  }
+  if (synctex == NULL) {
+    return true;
+  }
+
+  int line = 0;
+  int column = 0;
+  char* input_file = NULL;
+  if (synctex_parse_input(synctex, &input_file, &line, &column) == false) {
+    return false;
+  }
+
+  ret = synctex_view(zathura, input_file, line, column);
+  g_free(input_file);
+  return ret;
+}
+
 void
 document_open_idle(zathura_t* zathura, const char* path, const char* password,
-                   int page_number, const char* mode)
+                   int page_number, const char* mode, const char* synctex)
 {
-  if (zathura == NULL || path == NULL) {
-    return;
-  }
+  g_return_if_fail(zathura != NULL);
+  g_return_if_fail(path != NULL);
 
   zathura_document_info_t* document_info = g_try_malloc0(sizeof(zathura_document_info_t));
   if (document_info == NULL) {
@@ -908,10 +1143,17 @@ document_open_idle(zathura_t* zathura, const char* path, const char* password,
   }
 
   document_info->zathura     = zathura;
-  document_info->path        = path;
-  document_info->password    = password;
+  document_info->path        = g_strdup(path);
+  if (password != NULL) {
+    document_info->password  = g_strdup(password);
+  }
   document_info->page_number = page_number;
-  document_info->mode        = mode;
+  if (mode != NULL) {
+    document_info->mode      = g_strdup(mode);
+  }
+  if (synctex != NULL) {
+    document_info->synctex   = g_strdup(synctex);
+  }
 
   gdk_threads_add_idle(document_info_open, document_info);
 }
@@ -994,14 +1236,14 @@ document_close(zathura_t* zathura, bool keep_monitor)
   /* store file information */
   const char* path = zathura_document_get_path(zathura->document);
 
-  zathura_fileinfo_t file_info = { 0, 0, 1, 0, 1, 1, 0, 0 };
+  zathura_fileinfo_t file_info = { 0, 0, 1, 0, 1, "1:2", 0, 0 };
   file_info.current_page = zathura_document_get_current_page_number(zathura->document);
   file_info.page_offset  = zathura_document_get_page_offset(zathura->document);
   file_info.scale        = zathura_document_get_scale(zathura->document);
   file_info.rotation     = zathura_document_get_rotation(zathura->document);
 
   girara_setting_get(zathura->ui.session, "pages-per-row", &(file_info.pages_per_row));
-  girara_setting_get(zathura->ui.session, "first-page-column", &(file_info.first_page_column));
+  girara_setting_get(zathura->ui.session, "first-page-column", &(file_info.first_page_column_list));
 
   /* get position */
   file_info.position_x = zathura_document_get_position_x(zathura->document);
@@ -1015,6 +1257,9 @@ document_close(zathura_t* zathura, bool keep_monitor)
     zathura_db_save_jumplist(zathura->database, path, zathura->jumplist.list);
   }
 
+  /* free buffers */
+  g_free(file_info.first_page_column_list);
+
   girara_list_iterator_free(zathura->jumplist.cur);
   zathura->jumplist.cur = NULL;
   girara_list_free(zathura->jumplist.list);
@@ -1096,15 +1341,11 @@ statusbar_page_number_update(zathura_t* zathura)
     girara_setting_get(zathura->ui.session, "window-title-page", &page_number_in_window_title);
 
     if (page_number_in_window_title == true) {
-      bool basename_only = false;
-      girara_setting_get(zathura->ui.session, "window-title-basename", &basename_only);
-      char* title = g_strdup_printf("%s %s",
-        (basename_only == true)
-          ? zathura_document_get_basename(zathura->document)
-          : zathura_document_get_path(zathura->document),
-        page_number_text);
+      char* filename = get_formatted_filename(zathura, false);
+      char* title = g_strdup_printf("%s %s", filename, page_number_text);
       girara_set_window_title(zathura->ui.session, title);
       g_free(title);
+      g_free(filename);
     }
 
     g_free(page_number_text);
@@ -1244,10 +1485,10 @@ adjust_view(zathura_t* zathura)
 
   if (adjust_mode == ZATHURA_ADJUST_WIDTH ||
       (adjust_mode == ZATHURA_ADJUST_BESTFIT && page_ratio < view_ratio)) {
-    newscale = scale * (double)view_width / (double)document_width;
+    newscale *= (double)view_width / (double)document_width;
 
   } else if (adjust_mode == ZATHURA_ADJUST_BESTFIT) {
-    newscale = scale * (double)view_height / (double)cell_height;
+    newscale *= (double)view_height / (double)cell_height;
 
   } else {
     goto error_ret;
@@ -1259,8 +1500,8 @@ adjust_view(zathura_t* zathura)
   zathura_document_get_cell_size(zathura->document, &new_cell_height, &new_cell_width);
 
   /* if the change in scale changes page cell dimensions by at least one pixel, render */
-  if (abs(new_cell_width - cell_width) > 1 ||
-      abs(new_cell_height - cell_height) > 1) {
+  if (abs((int)new_cell_width - (int)cell_width) > 1 ||
+      abs((int)new_cell_height - (int)cell_height) > 1) {
     render_all(zathura);
     refresh_view(zathura);
 
@@ -1273,178 +1514,6 @@ error_ret:
   return false;
 }
 
-bool
-zathura_jumplist_has_previous(zathura_t* zathura)
-{
-  return girara_list_iterator_has_previous(zathura->jumplist.cur);
-}
-
-bool
-zathura_jumplist_has_next(zathura_t* zathura)
-{
-  return girara_list_iterator_has_next(zathura->jumplist.cur);
-}
-
-zathura_jump_t*
-zathura_jumplist_current(zathura_t* zathura)
-{
-  if (zathura->jumplist.cur != NULL) {
-    return girara_list_iterator_data(zathura->jumplist.cur);
-  } else {
-    return NULL;
-  }
-}
-
-void
-zathura_jumplist_forward(zathura_t* zathura)
-{
-  if (girara_list_iterator_has_next(zathura->jumplist.cur)) {
-    girara_list_iterator_next(zathura->jumplist.cur);
-  }
-}
-
-void
-zathura_jumplist_backward(zathura_t* zathura)
-{
-  if (girara_list_iterator_has_previous(zathura->jumplist.cur)) {
-    girara_list_iterator_previous(zathura->jumplist.cur);
-  }
-}
-
-static void
-zathura_jumplist_reset_current(zathura_t* zathura)
-{
-  g_return_if_fail(zathura != NULL && zathura->jumplist.cur != NULL);
-
-  while (true) {
-    if (girara_list_iterator_has_next(zathura->jumplist.cur) == false) {
-      return;
-    }
-
-    girara_list_iterator_next(zathura->jumplist.cur);
-  }
-}
-
-static void
-zathura_jumplist_append_jump(zathura_t* zathura)
-{
-  g_return_if_fail(zathura != NULL && zathura->jumplist.list != NULL);
-
-  zathura_jump_t* jump = g_try_malloc0(sizeof(zathura_jump_t));
-  if (jump == NULL) {
-    return;
-  }
-
-  jump->page = 0;
-  jump->x    = 0.0;
-  jump->y    = 0.0;
-  girara_list_append(zathura->jumplist.list, jump);
-
-  if (zathura->jumplist.size == 0) {
-    zathura->jumplist.cur = girara_list_iterator(zathura->jumplist.list);
-  }
-
-  ++zathura->jumplist.size;
-  zathura_jumplist_trim(zathura);
-}
-
-void
-zathura_jumplist_trim(zathura_t* zathura)
-{
-  g_return_if_fail(zathura != NULL && zathura->jumplist.list != NULL && zathura->jumplist.size != 0);
-
-  girara_list_iterator_t* cur = girara_list_iterator(zathura->jumplist.list);
-
-  while (zathura->jumplist.size > zathura->jumplist.max_size) {
-    if (girara_list_iterator_data(cur) == girara_list_iterator_data(zathura->jumplist.cur)) {
-      girara_list_iterator_free(zathura->jumplist.cur);
-      zathura->jumplist.cur = NULL;
-    }
-
-    girara_list_iterator_remove(cur);
-    --zathura->jumplist.size;
-  }
-
-  if (zathura->jumplist.size == 0 || (zathura->jumplist.size != 0 && zathura->jumplist.cur != NULL)) {
-    girara_list_iterator_free(cur);
-  } else {
-    zathura->jumplist.cur = cur;
-  }
-}
-
-void
-zathura_jumplist_add(zathura_t* zathura)
-{
-  g_return_if_fail(zathura != NULL && zathura->document != NULL && zathura->jumplist.list != NULL);
-
-  unsigned int pagenum = zathura_document_get_current_page_number(zathura->document);
-  double x = zathura_document_get_position_x(zathura->document);
-  double y = zathura_document_get_position_y(zathura->document);
-
-  if (zathura->jumplist.size != 0) {
-    zathura_jumplist_reset_current(zathura);
-
-    zathura_jump_t* cur = zathura_jumplist_current(zathura);
-
-    if (cur != NULL) {
-      if (cur->page == pagenum && cur->x == x && cur->y == y) {
-        return;
-      }
-    }
-  }
-
-  zathura_jumplist_append_jump(zathura);
-  zathura_jumplist_reset_current(zathura);
-  zathura_jumplist_save(zathura);
-}
-
-bool
-zathura_jumplist_load(zathura_t* zathura, const char* file)
-{
-  g_return_val_if_fail(zathura != NULL && file != NULL, false);
-
-  if (zathura->database == NULL) {
-    return false;
-  }
-
-  zathura->jumplist.list = zathura_db_load_jumplist(zathura->database, file);
-
-  if (zathura->jumplist.list == NULL) {
-    girara_error("Failed to load the jumplist from the database");
-
-    return false;
-  }
-
-  zathura->jumplist.size = girara_list_size(zathura->jumplist.list);
-
-  if (zathura->jumplist.size != 0) {
-    zathura->jumplist.cur = girara_list_iterator(zathura->jumplist.list);
-    zathura_jumplist_reset_current(zathura);
-    zathura_jumplist_trim(zathura);
-    girara_debug("Loaded the jumplist from the database");
-  } else {
-    girara_debug("No jumplist for this file in the database yet");
-  }
-
-  return true;
-}
-
-static void
-zathura_jumplist_save(zathura_t* zathura)
-{
-  g_return_if_fail(zathura != NULL && zathura->document != NULL);
-
-  zathura_jump_t* cur = zathura_jumplist_current(zathura);
-
-  unsigned int pagenum = zathura_document_get_current_page_number(zathura->document);
-
-  if (cur != NULL) {
-    cur->page = pagenum;
-    cur->x = zathura_document_get_position_x(zathura->document);
-    cur->y = zathura_document_get_position_y(zathura->document);
-  }
-}
-
 #ifdef G_OS_UNIX
 static gboolean
 zathura_signal_sigterm(gpointer data)
